#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#define _RS ' '                         /* Record Seperator */
#define VEC_TYPE long                   /* type used for BYG bit vector */
#define MAXBITS (8*sizeof(VEC_TYPE))    /* size in bits of BYG bit vector */
#define SIGMA 256                       /* bits per char */


typedef struct {
    void *addr;
    size_t size;
    int fd;
} mmap_t;
typedef mmap_t *Awale__TextSearch__MMap;

typedef struct {
    char *string;
    char skip[SIGMA];
} bm_t;
typedef bm_t *Awale__TextSearch__BM;


#ifndef DEBUG
#   define WARN(args) warn args
#else
#   define WARN(args) (void)(0)
#endif


/* convert mis-matches into bits to hold the info */
int k2bits[]={
 /* 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22*/
    1, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6
};

#define MAX_K (sizeof(k2bits)/sizeof(k2bits[0]))

/* ************************************************************* */
/* ERROR HANDLING ROUTINES */
/* we never actually use the last item - it's the NULL terminator */
#define ERRSTK 20
typedef struct {  
  char *msg;
  int size;
  int code;
} s_errmsg;
static s_errmsg error_stack[ERRSTK];
static int error_sp = ERRSTK - 1;

/* ************************************************************ */
void s_init_error(void)
{
    int q;
    for(q=0; q<ERRSTK; q++) {
        error_stack[q].code=0;
        error_stack[q].msg=NULL;
        error_stack[q].size=0;
    }
}

/* ************************************************************ */
void s_clear_error(void) 
{
    error_sp = ERRSTK-1;
}

/* ************************************************************ */
s_errmsg *s_errors(void)
{
    return &error_stack[error_sp];
}

/* ************************************************************ */
void s_push_error(int code, char const *msg) 
{
    s_errmsg *err;
    int size = strlen(msg)+1;

    if (error_sp <= 0)                    /* bad, bad programmer */
        return;

    --error_sp;
    err=&error_stack[error_sp];
    
    if (err->size < size) {
        if (err->msg != NULL)
            Safefree(err->msg);
        /* memory allocated on the following line is only ever release when
           we need a bigger string */
        Newc(123, err->msg, size, char, char);
        err->size=size;
    }

    memcpy(err->msg, msg, size);
    err->code=code;
}

/* ************************************************************* */
void s_push_errorvf(int code, char const *fmt, va_list ap) 
{
    char buf[1024];
#if defined(_MSC_VER)
    _vsnprintf(buf, sizeof(buf), fmt, ap);
#else
    /* is there a way to detect vsnprintf()?
        for this and other functions we need some mechanism to handle
        detection (like perl's Configure, or autoconf)
    */
    vsnprintf(buf, sizeof(buf), fmt, ap);
#endif
    s_push_error(code, buf);
}

/* ************************************************************* */
void s_push_errorf(int code, char const *fmt, ...) 
{
    va_list ap;
    va_start(ap, fmt);
    s_push_errorvf(code, fmt, ap);
    va_end(ap);
}



/* ************************************************************* */
/* ************************************************************* */

MODULE = Awale::TextSearch		PACKAGE = Awale::TextSearch		

PROTOTYPES: ENABLE

BOOT:
    s_init_error();


Awale::TextSearch::MMap
map_file(filename)
    char *filename;
  PREINIT:
    struct stat statbuf;
    int fd;
    mmap_t *addr;

  CODE:
    RETVAL=NULL;

    Newc(123, addr, 1, mmap_t, mmap_t);
    addr->addr=(void *)NULL;
    addr->size=0;
    addr->fd=-1;

    errno=0;
    if((fd=open(filename,O_RDONLY)) < 0) {
        s_push_errorf(errno, "Unable to open %s", filename);
    }
    else if(fstat(fd,&statbuf) <0) {
        s_push_errorf(errno, "Unable to stat %s", filename);
    }
    else if((addr->addr = mmap(0, statbuf.st_size, PROT_READ, 
                        MAP_FILE|MAP_SHARED, fd, 0)) == MAP_FAILED) {
        s_push_errorf(errno, "Unable to mmap %i bytes", statbuf.st_size);
    }
    else {
        addr->size=statbuf.st_size;
        addr->fd=fd;
        RETVAL=addr;
    }

    if(RETVAL == NULL) {            /* failed somewhere */
        if(fd != -1) 
            close(fd);                  /* clean up */
        if(addr != NULL) 
            Safefree(addr);
    }

  OUTPUT:
    RETVAL


void
map_dump(addr)
    Awale::TextSearch::MMap addr;
  CODE:
    warn("Awale::TextSearch::MMap={addr=%p, size=%i}", (void*)addr->addr,
                                                       (void*)addr->size);



    /* ******************************************************* */

void
linear_search(addr,word)
    Awale::TextSearch::MMap addr;
    char *word;
  PREINIT:
    int I, qf, qw, len;
    char *index;
  PPCODE:
    if(addr==NULL || addr->addr==NULL) 
        XSRETURN(0);

    len=strlen(word);
    if(len > addr->size)    /* edge case: search term is bigger'n whole index */
        XSRETURN(0);

    index=(char *)addr->addr;
    qw=0;                   /* index in word */
    I=0;                    /* pointer to start of word */
    if(word[0] == _RS)  
        I++;

    for(qf=0; qf < addr->size; qf++) {
        if(qw < len &&     /* if qw >= len, we are looking for next RS */
                    index[qf] == word[qw]) {
            WARN(("word_search: %c == %c\n", index[qf], word[qw]));
            qw++;
            if(qw >= len) {     /* found a word */
                WARN(("word_search: found %s word at offset %i\n", 
                                                (!(word[qw-1] == _RS) ? "stem" 
                                                   : "exact"), I));
                EXTEND(SP,1);   /* put it on the return stack */
                PUSHs(sv_2mortal(newSViv(I)));
            }
        }
        else if(index[qf] == _RS) {
            I=qf+1;
            qw=0;
        }
        else {
            qw=len;             /* failed, look for RS */
        }
    }


    /* ******************************************************* */

void
RS()
  PREINIT:
    char my_rs[2]={_RS, '\0'};
  PPCODE:
    EXTEND(SP,1);
    PUSHs(sv_2mortal(newSVpvn(my_rs, 1)));




Awale::TextSearch::BM
bm_study(pattern)
    char *pattern;
  PREINIT:
    bm_t *study;
    int q, M;
  CODE:
    Newc(123, study, 1, bm_t, bm_t);
    study->string=NULL;
    M=strlen(pattern);

    Newc(123, study->string, M+1, char, char);
    memcpy(study->string, pattern, M+1);

    RETVAL=study;

    /* default value for the entire skip table */
    for(q=0; q<SIGMA; q++) 
        study->skip[q]=M;

    /* chars in pattern have a lower skip value */
    for(q=0; q<M; q++)
        study->skip[(unsigned int)pattern[q]]=M-q-1;
    
  OUTPUT:
    RETVAL




void
bm_dump(study)
    Awale::TextSearch::BM study;
  PREINIT:
    int q;
  CODE:
    warn("Awale::TextSearch::BM\n");
    if(! (study->string == NULL)) {
        warn("M=%i, string='%s'\n", strlen(study->string), study->string);
    }
    else {
        warn("M=0, string=NULL\n");
    }
    for(q=0; q<sizeof(study->skip); q++)
        if(isprint((char)q))
            warn("'%c' = %i\n", (char)q, study->skip[q]);
        else 
            warn("%02x  = %i\n", q, study->skip[q]);

    /* ******************************************************* */


void
bm_search(addr, study)
    Awale::TextSearch::MMap addr;
    Awale::TextSearch::BM study;
  PREINIT:
    int i,j,t,M,N,I, ret=0;
    char *index;
  PPCODE:
    if(addr==NULL || addr->addr==NULL) 
        XSRETURN(0);

    M=strlen(study->string);
    N=addr->size;
    index=(char *)addr->addr;
    i=M-1;                          /* start compare from end */
    while(i < N) {
        for(j=M-1; j>=0; j--) {
            i--;
            while(study->string[j] != index[i]) {
                t=study->skip[(unsigned int)index[i]];
                WARN(("Missmatch at i=%i j=%i ('%c' != '%c'), t=%i\n", i, j,
                                    index[i], study->string[j], t));
                i+= (M-j > t) ? M-j : t;
                if(i>=N) {              /* end of index */
                    WARN(("End of index reached\n"));
                    XSRETURN(ret);
                }
                j = M-1;
            }
        }
        I=i+1;          /* past the RS */
        WARN(("Boyer-Moore match at %i\n", I));
        EXTEND(SP,1);   /* put it on the return stack */
        PUSHs(sv_2mortal(newSViv(I)));
        ret++;
        
        if(study->string[M]== _RS) {        /* exact search */
            i=N;
        }
        else {
            /* NB : pattern is assumed to start with RS */
            i+=M;                           /* skip over pattern */
            i+=M-1;                         /* start compare from end */
        }
    }
    WARN(("Last word matched\n"));

    /* ******************************************************* */

void
byg_search(addr,word,k)
    Awale::TextSearch::MMap addr;
    char *word;
    int k;

  PREINIT:
    int I, m, n, i, ret, bits;
    VEC_TYPE mask, ovmask, table[SIGMA], watch, state, ov;
    char *index;

  PPCODE:
    if(addr==NULL || addr->addr==NULL) 
        XSRETURN(0);

    index=(char *)addr->addr;
    m=strlen(word);
    n=addr->size;

    if(k==0 || m==1) {                  /* doing BYG would be overkill! */
        s_push_errorf(EINVAL, 
                    "Please do not do a BYG w/ k==0 or a word of 1 letter");
        XSRETURN(0);
    }

    if(m > MAXBITS) {                   /* to big */
        s_push_errorf(EINVAL, 
             "Please do not do a BYG search for something longer then %i", 
                            MAXBITS);
        XSRETURN(0);
    }

    if(m > n) {                         /* WAY to big */
        XSRETURN(0);
    }

    if(k > MAX_K) {                    /* sanity check */
        s_push_errorf(EINVAL,
                "%i mismatches is to many for this architecture, maxmium is %i", 
                        k, MAX_K);
        XSRETURN(0);
    }

    bits=k2bits[k];
    if(m * bits > MAXBITS) {            /* STILL to big */
        s_push_errorf(EINVAL, 
              "%i mismatches too many for a pattern %i long", k, m);
        XSRETURN(0);
    }

    /* initialise ovmask for masking out the counter overflow */    
    /* example, bits=3, ovmask is 0100  1001 0010 0100 */
    mask=1<<(bits-1);
    ovmask=0;
    for(i=0; i < m; i++) {
        ovmask |= mask;
        mask <<= bits;
    }

    /* init the counters so that all entries are inited to 1 */
    table[0]=ovmask >> (bits -1);
    for(i=1; i<SIGMA; i++) {
        table[i]=table[0];
    }

    /* the counters corresponding to the caracters of P are zeroed */
    /* note that mask now begins a new life */
    for(i=0, mask=1; i<m ; i++, mask <<= bits) {
        table[(unsigned int)word[i]] &= ~mask;
    }

    /* ******************************************************* */
    /* this is where the searching really starts */

    mask = (1 << (m * bits)) - 1;
    state = mask & ~ ovmask;
    ov = ovmask;
    watch = (k+1) << (bits * (m-1));
    ret=0;

    for(i=0; i < n; i++) {
        /* advance the state with the Shift and the And */
        state = ((state << bits) + table[(unsigned int)index[i]]) & mask;
        /* record any overflow */
        ov    = ((ov << bits) | (state & ovmask))   & mask;
        /* clear the overflow */
        state &= ~ ovmask;      
        if((state | ov) < watch) {      /* check for match */
            I=i-m+1;                    /* we have a match */
            if(word[0] == _RS)
                I++;
            WARN(("Baezer-Yates-Gonnet match at %i (i=%i, m=%i)\n", I, i, m));
            EXTEND(SP,1);
            PUSHs(sv_2mortal(newSViv(I)));
            ret++;
        }
    }
    WARN(("Found %i match(s)\n", ret));

    /* ******************************************************* */



void
clear_errors()
  CODE:
    s_clear_error();


void
search_errors()
  PREINIT:
    s_errmsg *errors;
    int i;
    HV *hv;
    SV *sv;

  PPCODE:
    errors=s_errors();
    for(i=0; errors[i].msg != NULL; i++) {
        hv=newHV();

        sv=newSVpv(errors[i].msg, strlen(errors[i].msg));
        if(!hv_store(hv, "msg", 3, sv, 0))
            SvREFCNT_dec(sv);

        sv = newSViv(errors[i].code);
        if(!hv_store(hv, "errno", 5, sv, 0))
            SvREFCNT_dec(sv);

        PUSHs(sv_2mortal(newRV_noinc((SV*)hv)));
    }

    /* ******************************************************* */



MODULE = Awale::TextSearch      PACKAGE = Awale::TextSearch::MMap

void
DESTROY(addr)
    Awale::TextSearch::MMap addr;
  CODE:
    WARN("MMap DESTROY\n");
    if(addr == NULL)
        XSRETURN(0);

    if(munmap(addr->addr, addr->size) == -1) {
        s_push_errorf(errno, "Unable to munmap %i bytes at %x", 
                                addr->size, addr->addr);
    }
    addr->addr=NULL;
    addr->size=0;
    
    if(addr->fd !=-1 && close(addr->fd)==1) {
        s_push_errorf(errno, "Unable to close file");
    }
    addr->fd=-1;
    Safefree(addr);




MODULE = Awale::TextSearch      PACKAGE = Awale::TextSearch::BM

void
DESTROY(study)
    Awale::TextSearch::BM study;
  CODE:
    WARN("BM DESTROY\n");
        XSRETURN(0);
    if(study->string!=NULL) 
        Safefree(study->string);

    if(study!=NULL) 
        Safefree(study);


